(*
  Call trees
*)

signature CALL_TREE =
sig
  datatype app_type = Rec | Constr | Def

  datatype tree =
    Variable of term * string
  | Case of term * Datatype.info * string * string list * tree * (string list * term * tree) list
  | If of term * string * string list * tree * tree * tree
  | App of term * app_type * string * tree list

  val i_term : tree -> term

  val add_defined_syms : tree -> string list -> string list
  val add_case_syms : tree -> string list -> string list
  val add_sim_syms : string -> tree -> string list -> string list

  val analyze_rules : string list -> (string -> bool) -> theory -> term list -> 
    (string * (term * tree) list) list

end

structure Call_Tree : CALL_TREE =
struct

open Isafor_Basics

(* explicit trees *)

datatype app_type = Rec | Constr | Def

datatype tree =
  Variable of term * string
| Case of term * Datatype.info * string * string list * tree * (string list * term * tree) list
| If of term * string * string list * tree * tree * tree
| App of term * app_type * string * tree list


fun fold_pats thy f t =
  let 
    val (Const (_, T), ts) = strip_comb t
    val elimT as (Type (dt, _)) = List.last (binder_types T)
    val constrs = the (Datatype.get_constrs thy dt)
    
    fun fold_pat ((cname, cT), lamterm) =
      let
        val n = length (binder_types cT)
        val (vs', b) = Term.strip_abs_eta n lamterm
        val pat = 
          list_comb (Const (cname, map snd vs' ---> elimT), map Free vs')
      in
        f (map fst vs') pat b
      end

    val (lams, _) = split_last ts
  in
    fold fold_pat (constrs ~~ lams)
  end



fun mk_tree _ _ _ _ _ (t as Free (v, _)) x = (Variable (t, v), x)
  | mk_tree f fs is_constr thy vs t i =
  let
    val (Const (n, _), ts) = strip_comb t
    fun new_name f i = f ^ "__" ^ string_of_int i
    fun extra_vars ts = fold Term.add_frees ts [] |> rev |> map fst
  in
    if n = @{const_name HOL.If} then
      let
        val [cond,thn,els] = ts
        val (cond_tree, i') = mk_tree f fs is_constr thy vs cond i
        val xs = extra_vars [thn, els]
        val (thn_tree, i'') = mk_tree f fs is_constr thy xs thn (i' + 1)
        val (els_tree, i''') = mk_tree f fs is_constr thy xs els i''
      in
        (If (t, new_name f i', xs, cond_tree, thn_tree, els_tree), i''')
      end
    else case Datatype.info_of_case thy n of
      NONE => 
        let val (ts', i') = fold_map (mk_tree f fs is_constr thy vs) ts i
            val app_type =
              if member (op =) fs n then Rec else if is_constr n then Constr else Def
        in (App (t, app_type, n, ts'), i') end
    | SOME dt =>
         let
           val (lams, arg) = split_last ts
           val (arg_tree, i') = mk_tree f fs is_constr thy vs arg i

           val xs = extra_vars lams

           fun add_subtree vs' pat b (i, sts) =
             let
               val (b_tree, i') = 
                 mk_tree f fs is_constr thy (vs' @ xs) b i
             in
               (i', (vs', pat, b_tree) :: sts)
             end
           val (i'', sts) = fold_pats thy add_subtree t (i' + 1, [])
          in
            (Case (t, dt, new_name f i', xs, arg_tree, rev sts), i'')
          end
    end

fun i_term (Case (_, _, f, extra, arg, _)) =  mk_app f (i_term arg :: map mk_var extra)
  | i_term (Variable (_, x)) = mk_var x
  | i_term (App (_, _, f, ts)) = mk_app f (map i_term ts)
  | i_term (If (_, f, extra, cond, _, _)) = mk_app f (i_term cond :: map mk_var extra)

(* FIXME: we do not need all defined symbols; only those that occur in arguments
   of recursive calls *)
fun add_defined_syms (Case (_, _, _, _, arg, children)) =
      add_defined_syms arg #> fold (add_defined_syms o #3) children
  | add_defined_syms (Variable _) = I
  | add_defined_syms (App (_, Def, f, ts)) =
      insert (op =) f #> fold add_defined_syms ts
  | add_defined_syms (App (_, _, _, ts)) =
      fold add_defined_syms ts
  | add_defined_syms (If (_, _, _, cond, thn, els)) =
      fold add_defined_syms [cond, thn, els]

fun contains_sym f (Case (_, _, _, _, arg, cs)) =
      contains_sym f arg orelse exists (contains_sym f o #3) cs
  | contains_sym _ (Variable _) = false
  | contains_sym f (App (_, _, g, ts)) = f = g orelse exists (contains_sym f) ts
  | contains_sym f (If (_, _, _, c, t, e)) = exists (contains_sym f) [c,t,e]

fun add_sim_syms f =
  let
    fun add b (Case (_, _, _, _, arg, cs)) =
          add (b orelse exists (contains_sym f o #3) cs) arg #> fold (add b o #3) cs
      | add _ (Variable _) = I
      | add b (App (_, Def, g, ts)) =
          (if b then insert (op =) g else I) #> fold (add b) ts
      | add b (App (_, apt, _, ts)) = fold (add (b orelse apt = Rec)) ts
      | add b (If (_, _, _, c, t, e)) =
          add (b orelse exists (contains_sym f) [t,e]) c #> fold (add b) [t,e]
  in add false end

fun add_case_syms (Case (_, _, f, _, arg, children)) =
      insert (op =) f #> add_case_syms arg #> fold (add_case_syms o #3) children
  | add_case_syms (Variable _) = I
  | add_case_syms (App (_, _, _, ts)) =
      fold add_case_syms ts
  | add_case_syms (If (_, f, _, cond, thn, els)) =
      insert (op =) f #> fold add_case_syms [cond, thn, els]

fun analyze_rules fs is_constr thy eqs =
  let
    fun case_rules (Case (_, _, f, extra, arg, children)) =
        (f, map (fn (_, p, tr) => 
        let val lhs = mk_app f (mk_term p :: map mk_var extra)
        in (lhs, tr) end) children)
      :: case_rules arg @ maps (case_rules o #3) children
      | case_rules (Variable _) = []
      | case_rules (App (_, _, _, trs)) = maps case_rules trs
      | case_rules (If (_, f, extra, cond, thn, els)) =
        (f, [(mk_app f (mk_app @{const_name HOL.True} [] :: map mk_var extra), thn),
             (mk_app f (mk_app @{const_name HOL.False} [] :: map mk_var extra), els)])
        :: maps case_rules [cond, thn, els]
    fun mk_rule eq i =
      let
        val (lhs, rhs) = HOLogic.dest_eq (HOLogic.dest_Trueprop eq)
        val vs = Term.add_frees lhs [] |> rev |> map fst
        val Const (f, _) = head_of lhs
        val (tr, i') = mk_tree f fs is_constr thy vs rhs i
       in ((f, (mk_term lhs, tr)), i') end
    val main_rules_trees = fst (fold_map mk_rule eqs 0)
    val trees = map (snd o snd) main_rules_trees
    val main_rules = AList.group (op =) main_rules_trees
    val more_rules = maps case_rules trees
  in
    main_rules @ more_rules
  end

end
